--[[
Name: LibSpreadsheet-1.0
Revision: $Rev: 34 $
Author(s): Humbedooh
Description: Spreadsheet stuff
Dependencies: LibStub AceGUI-3.0
License: GPL v3 or later.
]]
-----------------------------------------------------------------------
--dofile([[..\LibStub\LibStub.lua]]);
local SPREADSHEET = "LibSpreadsheet-1.0"
local SPREADSHEET_MINOR = tonumber(("$Rev: 034 $"):match("(%d+)")) or 9999;
if not LibStub then error(SPREADSHEET .. " requires LibStub.") end
LibSpreadsheet = LibStub:NewLibrary(SPREADSHEET, SPREADSHEET_MINOR)
if not LibSpreadsheet then return end

if ( table.wipe == nil ) then
	table.wipe = function(obj) obj = {}; end;
end
--[[
LibSpreadsheet-1.0 DEFAULTS
]]--
LibSpreadsheet.disabled = LibSpreadsheet.disabled or nil;
LibSpreadsheet.objects = LibSpreadsheet.objects or {};
LibSpreadsheet.cols = {'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z', '?'};
LibSpreadsheet.sortAscending = 1;
LibSpreadsheet.sortDescending = 2;
LibSpreadsheet.sortAscendingWithCase = 3;
LibSpreadsheet.sortDescendingWithCase = 4;
LibSpreadsheet.callbackEvents = { 'OnSelectionChange', 'OnUserInputProcessed', 'OnUserInputRequested', 'OnUserSortRequested', 'OnUserSortProcessed' }

function LibSpreadsheet:GenerateRow(heap)
	local struct = {
		c = {1,1,1,1},
		b = {0,0,0,0},
		f = nil,
		h = false,
		u = true,
		cells = {}
	};
	for key, col in pairs(self.columns or {}) do
		table.insert(struct.cells,
		{
			f = nil,
			c = nil,
			b = nil,
            id = nil,
			data = nil,
			u = true,
			w = 0,
			h = false
		}
		);
	end
	table.insert(heap, struct);
	return heap[#heap];
end

function LibSpreadsheet:DeregisterRow(row)
    if ( row.frame ) then
        for k, cell in pairs(row.cells) do
            if ( cell.frame ~= nil ) then
                cell.frame:Hide();
                cell.frame:SetParent(nil);
                cell.frame = nil;
            end
        end
        row.frame:Hide();
        row.frame:SetParent(nil);
        row.frame = nil;
    end
end



local function deepcopy(object)
	local lookup_table = {}
	local function _copy(object)
		if type(object) ~= "table" then
			return object
		elseif lookup_table[object] then
			return lookup_table[object]
		end
		local new_table = {}
		lookup_table[object] = new_table
		for index, value in pairs(object) do
			new_table[_copy(index)] = _copy(value)
		end
		return setmetatable(new_table, getmetatable(object))
	end
	return _copy(object)
end

--[[
LibSpreadsheet-1.0 BOOK API
]]--

function LibSpreadsheet:SetupSheet(sheet, book)
    local xbook = book;
    sheet.parent =                    function() return xbook; end;
    sheet.rowHeap =                   {}
    sheet.heapChanged =               true;
    sheet.frame =                     nil;
    sheet.header =                    nil;
	sheet.GenerateHeap =              LibSpreadsheet.GenerateHeap;
    sheet.GenerateRow  =              LibSpreadsheet.GenerateRow;
    sheet.DeregisterRow =             LibSpreadsheet.DeregisterRow;
    sheet.AddColumn =                 LibSpreadsheet.AddColumn;
    sheet.SetColumn =                 LibSpreadsheet.SetColumn;
    sheet.DeleteColumn =              LibSpreadsheet.DeleteColumn;
    sheet.ExtractLocation =           LibSpreadsheet.ExtractLocation;
    sheet.AddRow =                    LibSpreadsheet.AddRow;
    sheet.DeleteRow =                 LibSpreadsheet.DeleteRow;
    sheet.GoTo =                      LibSpreadsheet.GoTo;
    sheet.Insert =                    LibSpreadsheet.Insert;
    sheet.Delete =                    LibSpreadsheet.Delete;
    sheet.SetCellValue =              LibSpreadsheet.SetCellValue;
    sheet.StartSelect =               LibSpreadsheet.StartSelect;
    sheet.EndSelect =                 LibSpreadsheet.EndSelect;
    sheet.SetVisibleSelection =       LibSpreadsheet.SetVisibleSelection;
    sheet.UpdateSelect =              LibSpreadsheet.UpdateSelect;
    sheet.SetInteractive =            LibSpreadsheet.SetInteractive;
    sheet.SetColorRange =             LibSpreadsheet.SetColorRange;
    sheet.SetRow =                    LibSpreadsheet.SetRow;
    sheet.ClearCells =                LibSpreadsheet.ClearCells;
    sheet.SortByColumn =              LibSpreadsheet.SortByColumn;
    sheet.SetCallback =               LibSpreadsheet.SetSheetCallback;
    sheet.UpdateFrameSizes =          LibSpreadsheet.UpdateFrameSizes;
    sheet.ShowEditBox =               LibSpreadsheet.ShowEditBox;
    sheet.SaveEditBox =               LibSpreadsheet.SaveEditBox;
    sheet.SetCellSpacing =            LibSpreadsheet.SetCellSpacing;
    sheet.SetCellPadding =            LibSpreadsheet.SetCellPadding;
    sheet.GroupBy =                   LibSpreadsheet.GroupByColumn;
    sheet.UpdateHeap =                LibSpreadsheet.UpdateHeap;
    sheet.GetUniqueValues =           LibSpreadsheet.GetUniqueValues;
    sheet.AddFilter =                 LibSpreadsheet.AddFilter;
    sheet.ClearFilters =              LibSpreadsheet.ClearFilters;
    sheet.HideColumn =                LibSpreadsheet.HideColumn;
    sheet.ShowColumn =                LibSpreadsheet.ShowColumn;
    sheet.Commit =                    LibSpreadsheet.Commit;
    for k, v in pairs (sheet.columns) do
        v.frame = nil;
    end
    for k, v in pairs (sheet.rows) do
        v.frame = nil;
        for k, c in pairs (v.cells) do
            c.frame = nil;
        end
    end
end

function LibSpreadsheet:Load(obj)
    obj.Add =               LibSpreadsheet.AddSheet;
	obj.Delete =            LibSpreadsheet.DeleteSheet;
	obj.Select =            LibSpreadsheet.Select;
	obj.StrictMode =        LibSpreadsheet.StrictMode;
    obj.ShowSelector =      LibSpreadsheet.ShowSelector;
    obj.ShowScrollbars =    LibSpreadsheet.ShowScrollbars;
    obj.Leaf =              LibSpreadsheet.Leaf;
    obj.Render =            LibSpreadsheet.Render;
    obj.Hide =              LibSpreadsheet.Hide;
    obj.BuildFramework =    LibSpreadsheet.BuildFramework;
    obj.BuildFontStrings =  LibSpreadsheet.BuildFontStrings;
    obj.SetFont =           LibSpreadsheet.SetFont;
    obj.frame =             nil;
    obj.hasBookUI =         false;
    obj.hasSelector =       false;
    obj.editbox =           nil;
    for k, sheet in pairs(obj.sheets) do
        LibSpreadsheet:SetupSheet(sheet, obj);
    end
end

function LibSpreadsheet:Book(title)
	local obj = {};
	if ( type(title) ~= "string" ) then error(("%s: Error in creating book title; Expected string, got %s!"):format(SPREADSHEET, type(title))); end

    -- Book default values
	obj.sheets =            {};
    obj.parent =            self;
    obj.selectedSheet =     nil;
	obj.type =              'book';
	obj.title =             title;
    obj.strict =            false;
    obj.hasBookUI =         false;
    obj.hasSelector =       false;
    obj.editbox =           nil;

    -- Book functions
	obj.Add =               LibSpreadsheet.AddSheet;
	obj.Delete =            LibSpreadsheet.DeleteSheet;
	obj.Select =            LibSpreadsheet.Select;
	obj.StrictMode =        LibSpreadsheet.StrictMode;
    obj.ShowSelector =      LibSpreadsheet.ShowSelector;
    obj.ShowScrollbars =    LibSpreadsheet.ShowScrollbars;
    obj.Leaf =              LibSpreadsheet.Leaf;
    obj.Render =            LibSpreadsheet.Render;
    obj.Hide =              LibSpreadsheet.Hide;
    obj.BuildFramework =    LibSpreadsheet.BuildFramework;
    obj.BuildFontStrings =  LibSpreadsheet.BuildFontStrings;
    obj.SetFont =           LibSpreadsheet.SetFont;
    obj.headerFont =        nil;
    obj.numberFont =        nil;
    obj.stringFont =        nil;
	return obj;
end


--[[
LibSpreadsheet-1.0 SHEET API
]]--

function LibSpreadsheet:AddSheet(title)
	if ( type(title) ~= "string" ) then error(("%s: Error in setting sheet title; Expected string, got %s!"):format(SPREADSHEET, type(title))); end
	for k, v in pairs(self.sheets) do
		if ( v.title == title ) then error(("%s: Sheet \"%s\" already exists in book \"%s\"!"):format(SPREADSHEET, title, self.title)); end
	end
	local sheet = {
        parent =                    obj,
		type =                      'sheet',
		title =                     title,
        callbacks =                 {},
		columns =                   {},
        sortDirection =             LibSpreadsheet.sortAscending,
        lastColumn =                0,
		rows =                      {},
        groups =                    {},
        filters =                   {},
        hasGroups =                 false,
        hasFilters =                false,
        heapChanged =               false,
		rowHeight =                 0,
		lastRow =                   0,
		lastColumn =                0,
        currentRow =                nil,
        currentColumn =             nil,
		frame =                     nil,
        isInteractive =             false,
        cellSpacing =               2,
        cellPadding =               2,
	}
    LibSpreadsheet:SetupSheet(sheet, self);
	table.insert(self.sheets, sheet);
	return self:Select(sheet.title);
end

function LibSpreadsheet:DeleteSheet(title)
    local i = nil;
	if ( type(title) ~= "string" ) then error(("%s: Error in deleting sheet; Invalid sheet name - expected string, got %s!"):format(SPREADSHEET, type(title))); end
	for k, v in pairs(self.sheets) do
		if ( v.title == title ) then
            if ( self.selectedSheet == v ) then
                self.selectedSheet = nil;
            end
			i = k;
			break;
		end
	end
    if ( i~= nil ) then
        if ( self.sheets[i].frame ~= nil ) then
            self.sheets[i].frame:Hide();
            self.sheets[i].frame:SetParent(nil); -- make sure the garbage collector frees up this frame and its children
            self.sheets[i].header:Hide();
            self.sheets[i].header:SetParent(nil);
        end
        table.remove(self.sheets, i);
    end
end

function LibSpreadsheet:Select(sheet)
	for k, v in pairs(self.sheets) do
		if ( v.title == sheet ) then
            self.selectedSheet = v;
            if ( self.hasBookUI == true ) then
                self.verticalScrollbar:SetValue(0);
            end
			return v;
		end
	end
	return nil;
end


--[[
LibSpreadsheet-1.0 HEADER FUNCTIONS
]]--

function LibSpreadsheet:AddColumn(...)
	for i = 1, select("#", ...) do
		local column = select(i, ...);
        if ( type(column) == "table" or type(column) == "string" ) then
            local col = {
                name =              "",
                u =                 true,
                width =             1,
                height =            1,
                c =                 {1,1,1,1},
                b =                 {0,0,0,0},
                hidden =            false,
            };
            table.insert(self.columns, col);
            if ( self:SetColumn(#self.columns, column) == false ) then
                table.remove(#self.columns);
            else
                self.lastColumn = #self.columns;
            end
        else
            error(("%s: Error while adding column, expected string or table, got %s!"):format(SPREADSHEET, column));
        end
	end
    self.heapChanged = true;
end

function LibSpreadsheet:SetColumn(column, properties)
    local c = self:ExtractLocation(column);
    if ( self.columns[c] == nil ) then error(("%s: Error while setting column %s: No such column!"):format(SPREADSHEET, LibSpreadsheet.cols[c])); end
    if ( self.columns[c].type == nil or self.columns[c].justify == nil ) then
        self.columns[c].type = "string";
        self.columns[c].justify = "LEFT";
    end
    if ( type(properties) == "string" ) then
        self.columns[c].name=properties;
        self.columns[c].u = true;
        self.heapChanged = true;
        return true;
    elseif ( type(properties) == "table" ) then
        for key, val in pairs(properties) do
            if ( key == "color" ) then key = "c"; end
            if ( key == "background" ) then key = "b"; end
            self.columns[c][key] = val;
            if ( key == "type" and properties.justify == nil ) then
                if ( val == "number" ) then
                    self.columns[c].justify = "RIGHT";
                else
                    self.columns[c].justify = "LEFT";
                end
            end
        end
        self.columns[c].u = true;
        self.heapChanged = true;
        return true;
    else
        error(("%s: Error while setting column properties, expected string or table, got %s!"):format(SPREADSHEET, type(properties)));
        return false;
    end
end


function LibSpreadsheet:DeleteColumn(column)
    local c = self:ExtractLocation(column);
    if ( self.columns[c] == nil ) then error(("%s: Error while deleting column %s: No such column!"):format(SPREADSHEET, LibSpreadsheet.cols[c])); end
    table.remove(self.columns, c);
    self.lastColumn = #self.columns;
    for k, v in pairs(self.rows) do
        if ( v.cells[c] ~= nil ) then
            if ( v.cells[c].frame ~= nil ) then
                 v.cells[c].frame:Hide();
                 v.cells[c].frame:SetParent(nil);
             end
         end
        table.remove(v.cells, c);
    end
    self.heapChanged = true;
end



--[[
LibSpreadsheet-1.0 DATA FUNCTIONS
]]--

function LibSpreadsheet:Commit()
    if ( table.maxn(self.rowHeap) > 0 ) then
        self:ClearCells(false);
        for key, row in pairs(self.rowHeap) do
            if ( row.frame ) then
                for k, cell in pairs(row.cells) do
                    if ( cell.frame ) then
                        cell.frame:Hide();
                        cell.frame:SetParent(nil);
                        cell.frame = nil;
                    end
                end
                row.frame:Hide();
                row.frame:SetParent(nil);
                row.frame = nil;
            end
        end
        self.rows = self.rowHeap;
        self.rowHeap = {};
    end
end

function LibSpreadsheet:ClearCells(toggle)
    table.wipe(self.rows);
    self.hasFilters = false;
    self.hasGroups = false;
    if ( self.frame ~= nil ) then
        self.frame:Hide();
        self.frame:SetParent(nil);
        self.frame = nil;
    end
    if ( toggle ~= false ) then
        self.rowHeap = {};
    end
    self.heapChanged = true;
end

function LibSpreadsheet:GetUniqueValues(col)
    local tmp = {};
    local ret = {};
    local column = self:ExtractLocation(col);
    for k, row in pairs (self.rows) do
        if ( row.cells[column].data ~= nil ) then
            if ( tmp[row.cells[column].data] == nil ) then
                tmp[row.cells[column].data] = 1;
                table.insert(ret, row.cells[column].data);
            end
        end
    end
    return ret;
end

function LibSpreadsheet:GroupByColumn(...)
    table.wipe(self.groups);
    for i = 1, select("#", ...) do
		local column = self:ExtractLocation(select(i, ...));
        if ( self.columns[column] ~= nil ) then
            table.insert(self.groups, column);
        end
    end
    if ( table.maxn(self.groups) > 0 ) then
        self.hasGroups = true;
    else
        self.hasGroups = false;
    end
    self.heapChanged = true;
end

function LibSpreadsheet:ClearFilters()
    table.wipe(self.filters);
    self.heapChanged = true;
    self.hasFilters = false;
end

function LibSpreadsheet:AddFilter(...)
    for i = 1, select("#", ...) do
        local entry = select(i, ...);
		local column = self:ExtractLocation(entry[1]);
        if ( self.columns[column] ~= nil ) then
            table.insert(self.filters, {column, entry[2], entry[3]});
        end
    end
    self.hasFilters = true;
    self.heapChanged = true;
end


function LibSpreadsheet:SortByColumn(...)
    local func = function() return false; end;
    local i = select("#", ...);
    local p = 0;
    local N = 0;
    local rowTable = self.rows;
    self.heapChanged = true;
    if ( self.hasFilters or self.hasGroups ) then
        self:UpdateHeap();
        rowTable = self.rowHeap;
        self.heapChanged = false;
    end
    while ( i > p ) do
        local col = self:ExtractLocation(select((i-p-1), ...));
        local dir = select((i-p), ...);
        local l = string.lower;
        if      ( dir == LibSpreadsheet.sortAscending )         then dir = false;
        elseif  ( dir == LibSpreadsheet.sortAscendingWithCase ) then dir = false; l = function(x) return x; end;
        elseif  ( dir == LibSpreadsheet.sortDescending )        then dir = true;
        else                                                         dir = true;  l = function(x) return x; end;
        end
        p = p + 2;
        N = N + 1;
        local n = N;
        local r = dir;
        local c = col;
        if ( self.columns[c].type == "number" ) then
            l = function(x) return x; end;
        end
        local myFunc = func;
        func = function(a,b) if ( a == nil or b == nil ) then return; end if (l(a.cells[c].data) or 0) > (l(b.cells[c].data) or 0) then return r; elseif (l(a.cells[c].data) or 0) == (l(b.cells[c].data) or 0) then return myFunc(a,b); end return not r; end;
    end
    self.sortDirection = math.fmod(self.sortDirection, 2) + 1;
    table.sort(rowTable, func);
    for key, row in pairs(rowTable) do
        row.u = true;
    end
end


function LibSpreadsheet:AddRow(...)
    local row = self:GenerateRow(self.rows);
    for i = 1, select("#", ...) do
        local arg = select(i, ...);
        if ( type(arg) ~= "table" ) then
            arg = {data=arg};
        end
        row.cells[i].data = arg.data;
        row.cells[i].c = arg.color;
        row.cells[i].b = arg.background;
        row.cells[i].id = arg.id;
        if ( self.parent().strict == true ) then
            local t =  self.columns[i].type;
            if ( t == "string" and type(arg.data) ~= "string" ) then error(("%s: Error while adding cell in column %s, expected string, got %s!"):format(SPREADSHEET, LibSpreadsheet.cols[i], type(arg.data))); end
            if ( t == "boolean" and type(arg.data) ~= "boolean" ) then error(("%s: Error while adding cell in column %s, expected boolean, got %s!"):format(SPREADSHEET, LibSpreadsheet.cols[i], type(arg.data))); end
            if ( ( t == "number" ) and type(arg.data)  ~= "number" ) then error(("%s: Error while adding cell in column %s, expected number, got %s!"):format(SPREADSHEET, LibSpreadsheet.cols[i], type(arg.data))); end
        end
    end
    self.heapChanged = true;
    return #self.rows;
end



function LibSpreadsheet:SetRow(row, properties)
    row = tonumber(row);
    if ( self.rows[row] == nil ) then error(("%s: Error while setting row properties %s: No such row!"):format(SPREADSHEET, LibSpreadsheet.cols[c])); end
    if ( type(properties) == "table" ) then
        for key, val in pairs(properties) do
            self.rows[row][key] = val;
        end
        for k, v in pairs(self.rows[row].cells) do
            v.u = true;
        end
        self.rows[row].u = true;
    end
    self.heapChanged = true;
end

function LibSpreadsheet:DeleteRow(...)
    local dels = {};
    for i = 1, select("#", ...) do
		local row = select(i, ...);
        if ( self.rows[row] ~= nil ) then
            table.insert(dels, row);
            self:DeregisterRow(row);
        end
    end
    table.sort(dels, function(a,b) return (a > b); end);
    for k, v in pairs(dels) do
        table.remove(self.rows, v);
    end
    for k, row in pairs(self.rows) do
        row.positionChanged = true;
    end
    self.heapChanged = true;
end

function LibSpreadsheet:SetCellValue(c, r, d)
    local column, row = self:ExtractLocation(c);
    if ( row ~= nil ) then
        d = r;
    else
        row = r;
    end
    if ( self.rows[row] ~= nil ) then
        if ( self.rows[row].cells[column] ~= nil ) then
            self.rows[row].cells[column].data = d;
            self.rows[row].cells[column].u = true;
        end
    end
    self.heapChanged = true;
end

function LibSpreadsheet:ExtractLocation(str)
    local col = nil;
    local row = nil;
    if ( type(str) == "string" ) then
        local c, r = (str):match("(%D+)(%d*)");
        if ( r ~= nil ) then
            row = r;
        end
        if ( string.len(c) == 1 ) then
            col = tonumber(c, 36) - 9;
        else
            for k, column in pairs(self.columns) do
                if ( column.name == c ) then
                    col = k;
                    break;
                end
            end
        end
    elseif ( type(str) == "number" ) then
        if ( str > 26 ) then str = math.fmod(str, 26); end
        if ( str < 1 ) then
            while ( str < 1 ) do
                str = str + #self.columns;
            end
        end
        col = str;
    end
    return col, tonumber(row);
end



function LibSpreadsheet:GetValue(column, row)
	local c, r = self:ExtractLocation(column);
    if ( type(r) == "number" ) then
        row = r;
    end
    if ( self.columns[column] == nil ) then error(("%s: Error in GetValue(): No such column '%s'!"):format(SPREADSHEET, LibSpreadsheet.cols[column or 27])); end
    if ( self.rows[row] == nil ) then error(("%s: Error in GetValue(): No such row '%u'!"):format(SPREADSHEET, row)); end
    return self.rows[row].cells[column].data;
end

function LibSpreadsheet:Delete(fromCol, fromRow, toCol, toRow)
    local coords = {};
    if ( type(fromCol) == "number" ) then
        coords.selectStartColumn = fromCol;
        coords.selectStartRow = fromRow;
    elseif ( type(fromCol) == "string" ) then
        local c, r = self:ExtractLocation(fromCol);
        if ( type(r) == "number" ) then
            coords.selectStartColumn = c;
            coords.selectStartRow = r;
            toRow = toCol;
            toCol = fromRow;
        else
            coords.selectStartColumn = c;
            coords.selectStartRow = fromRow;
        end
    end
    if ( type(toCol) == "number" ) then
        coords.selectEndColumn = toCol;
        coords.selectEndRow = toRow;
    elseif ( type(toCol) == "string" ) then
        local c, r = self:ExtractLocation(toCol);
        if ( type(r) == "number" ) then
            coords.selectEndColumn = c;
            coords.selectEndRow = r;
        else
            coords.selectEndColumn = c;
            coords.selectEndRow = toRow;
        end
    end
    for x = math.min(coords.selectStartRow, coords.selectEndRow), math.max(coords.selectStartRow, coords.selectEndRow) do
        for y = math.min(coords.selectStartColumn, coords.selectEndColumn), math.max(coords.selectStartColumn, coords.selectEndColumn) do
            self.rows[x].cells[y].data = nil;
        end
    end
    self.heapChanged = true;
end

function LibSpreadsheet:Insert(...)

end

function LibSpreadsheet:StrictMode(toggle)
	if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting strict mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
	self.strict = toggle;
end

function LibSpreadsheet:UpdateHeap()
    if ( self.heapChanged ) then
        for key, row in pairs(self.rowHeap) do
            self:DeregisterRow(row);
        end
        local rowTable = self.rows;
        if ( self.hasFilters ) then
            for key, entry in pairs(self.filters) do
                local tmp = {};
                local col = entry[1];
                local method = entry[2];
                local val = entry[3];
                for key, row in pairs(rowTable) do
                    if ( method == "EQUAL" ) then
                        if ( row.cells[col].data == val ) then
                            local r = deepcopy(row);
                            r.frame = nil;
                            r.u = true;
                            for k, v in pairs(r.cells) do v.frame = nil; v.u=true; end
                            table.insert(tmp, r);
                        end
                    elseif ( method == "GREATER" ) then
                        if ( row.cells[col].data > val ) then
                            local r = deepcopy(row);
                            r.frame = nil;
                            r.u = true;
                            for k, v in pairs(r.cells) do v.frame = nil; v.u=true; end
                            table.insert(tmp, r);
                        end
                    elseif ( method == "LESS" ) then
                        if ( row.cells[col].data < val ) then
                            local r = deepcopy(row);
                            r.frame = nil;
                            r.u = true;
                            for k, v in pairs(r.cells) do v.frame = nil; v.u=true; end
                            table.insert(tmp, r);
                        end
                    else
                        if ( row.cells[col].data ~= val ) then
                            local r = deepcopy(row);
                            r.frame = nil;
                            r.u = true;
                            for k, v in pairs(r.cells) do v.frame = nil; v.u=true; end
                            table.insert(tmp, r);
                        end
                    end
                end
                rowTable = tmp;
            end
        end
        if ( self.hasGroups ) then
            for key, col in pairs(self.groups) do
                local tmp = {};
                for key, row in pairs(rowTable) do
                    if ( tmp[row.cells[col].data] ~= nil ) then
                        for key, cell in pairs(row.cells) do
                            if ( self.columns[key].type == "number" and key ~= column ) then
                                tmp[row.cells[col].data].cells[key].data = math.max(self.columns[key].minval or -math.huge, math.min(self.columns[key].maxval or math.huge, (tmp[row.cells[col].data].cells[key].data or 0) + (cell.data or 0)));
                            end
                        end
                    else
                        tmp[row.cells[col].data] = {cells={}};
                        for k, cell in pairs(row.cells) do
                            tmp[row.cells[col].data].cells[k] = deepcopy(cell);
                            tmp[row.cells[col].data].cells[k].frame = nil;
                            tmp[row.cells[col].data].cells[k].u = true;
                        end
                    end
                end
                local ntmp = {};
                for k, v in pairs(tmp) do
                    table.insert(ntmp, v);
                end
                rowTable = ntmp;
            end
        end
        self.rowHeap = rowTable;
        self.heapChanged = false;
    end
end



--[[
LibSpreadsheet-1.0 VISUAL HANDLERS
]]--


function LibSpreadsheet:ShowSelector(toggle)
    if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting selector mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
	self.showSelector = toggle;
end

function LibSpreadsheet:ShowScrollbars(toggle)
    if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting scrollbar visibility; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
	self.showScrollbars = toggle;
end

function LibSpreadsheet:SetInteractive(toggle)
    if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting interactive mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
    self.isInteractive = toggle;
end

function LibSpreadsheet:UpdateFrameSizes()
    local rowTable = self.rows;
    if ( self.hasGroups or self.hasFilters ) then
        self:UpdateHeap();
        rowTable = self.rowHeap;
    end
    for key, header in pairs (self.columns) do
        if ( header.u ) then
            self.parent().str:SetFontObject(self.parent().headerFont);
            self.parent().str:SetText((header.name or "??") .. ":");
            local height = math.ceil(self.parent().str:GetStringHeight());
            local width = math.ceil(self.parent().str:GetStringWidth());
            if ( header.h ~= height ) then header.h = height; end
            if ( header.w ~= width ) then header.w = width; end
            self.rowHeight = height + (self.cellPadding*2);
            header.colwidth = width + (self.cellPadding*2);
        end
    end
    local t = nil;
    for key, row in pairs (rowTable) do
        for key, cell in pairs (row.cells) do
            t = self.columns[key].type;
            if ( cell.u ) then
                if ( t == "number" ) then
                    self.parent().str:SetFontObject(self.parent().numberFont);
                    cell.data = tonumber(cell.data);
                    local s = ("%%.%uf%%s"):format(self.columns[key].decimals or 0);
                    self.parent().str:SetText(string.format(s, cell.data or 0, (self.columns[key].suffix or "")));
                else
                    self.parent().str:SetFontObject(self.parent().stringFont);
                    self.parent().str:SetText(cell.data or "");
                end
                cell.h = math.ceil(self.parent().str:GetStringHeight());
                cell.w = math.ceil(self.parent().str:GetStringWidth() + (self.cellPadding*2));
            end
            if ( cell.w > self.columns[key].colwidth ) then
                self.columns[key].colwidth = cell.w;
            end
            if ( cell.h > self.rowHeight ) then
                self.rowHeight = cell.h;
            end
        end
    end
end

function LibSpreadsheet:BuildFontStrings()
    self.str = UIParent:CreateFontString();
    local garble = string.format("%8x", math.random(time()));
    self.headerFont = CreateFont("LibSpreadsheetHeaderFont"..garble);
    self.numberFont = CreateFont("LibSpreadsheetNumberFont"..garble);
    self.stringFont = CreateFont("LibSpreadsheetStringFont"..garble);
    local locale =            (GetLocale or (function() return "enUS"; end))();
    self.numberFont:SetFont(    [[Fonts\ARIALN.TTF]],   12);
    if ( locale == "zhCN" ) then
        self.headerFont:SetFont([[Fonts\ZYKai_T.TTF]],  12);
        self.stringFont:SetFont([[Fonts\ZYKai_T.TTF]],  10);
    elseif ( locale == "zhTW" ) then
        self.headerFont:SetFont([[Fonts\bLEI00D.TTF]],  12);
        self.stringFont:SetFont([[Fonts\bLEI00D.TTF]],  10);
    elseif ( locale == "koKR" ) then
        self.headerFont:SetFont([[Fonts\2002.TTF]],     12);
        self.stringFont:SetFont([[Fonts\2002.TTF]],     10);
    else
        self.headerFont:SetFont([[Fonts\FRIZQT__.TTF]], 12);
        self.stringFont:SetFont([[Fonts\FRIZQT__.TTF]], 10);
    end
end

function LibSpreadsheet:SetFont(what, font, size, flags)
    what = string.lower(what);
    local garble = string.format("%8x", math.random(time()));
    local fontstring = CreateFont("LibSpreadsheetCustomFont"..garble);
    fontstring:SetFont(font, size, flags);
    if ( what == "header") then self.headerFont = fontstring end
    if ( what == "string") then self.headerFont = fontstring end
    if ( what == "number") then self.headerFont = fontstring end
end

function LibSpreadsheet:BuildFramework()
    if ( self.str == nil ) then
        self:BuildFontStrings();
    end
    self.frame = CreateFrame("Frame", nil, UIParent);
    self.outerScrollframe = CreateFrame("ScrollFrame", nil, self.frame);
    self.header = nil;
    self.frame:SetPoint("CENTER", UIParent, "CENTER", 0,0);
    self.frame:SetFrameStrata("DIALOG")
    self.frame:EnableMouse(true)
    self.innerFrame = nil;
    self.frame:Hide();
    self.offsetY = 0;
    self.scrollFrame = CreateFrame("ScrollFrame", nil, self.frame);
	self.scrollFrame:SetPoint("TOPLEFT");
	self.scrollFrame:SetPoint("BOTTOMRIGHT");
	self.scrollFrame:EnableMouseWheel(true)
    self.scrollFrame:EnableMouse(true)
    self.scrollFrame:SetAllPoints();

    self.verticalScrollbar = CreateFrame("Slider", nil, self.scrollFrame, "UIPanelScrollBarTemplate");
	self.verticalScrollbar:SetPoint("TOPLEFT", self.scrollFrame, "TOPRIGHT", 4, -16)
	self.verticalScrollbar:SetPoint("BOTTOMLEFT", self.scrollFrame, "BOTTOMRIGHT", 4, 16)
	self.verticalScrollbar:SetMinMaxValues(0, 1000)
	self.verticalScrollbar:SetValueStep(1)
	self.verticalScrollbar:SetValue(0)
	self.verticalScrollbar:SetWidth(16)
    self.scrollFrame.scrollvalue = 0;
	self.verticalScrollbar:Hide()
	self.verticalScrollbar:SetScript("OnValueChanged", function(a,b) self.scrollFrame:scrollFunc(b); end );

    self.scrollFrame.onMouseWheel = function(caller, value)
		local height, viewheight = self.scrollFrame:GetHeight(), self.innerFrame:GetHeight()

		if height > viewheight then
			self.verticalScrollbar:Hide()
		else
            self.verticalScrollbar:Show()
			local diff = height - viewheight
			local delta = 1
			if (value or 0) < 0 then
				delta = -1
			end
			self.verticalScrollbar:SetValue(min(max(self.scrollFrame.scrollvalue + delta*(self.offsetY/(diff/45)),0), 1000))
		end
	end;


    self.sheetSelector = CreateFrame("Frame", nil, self.frame);
    self.sheetSelector:SetFrameStrata("FULLSCREEN_DIALOG")
    self.prevSheet = CreateFrame("Button", nil, self.sheetSelector);
    self.prevSheet:SetWidth(32);
    self.prevSheet:SetHeight(32);
    self.prevSheet:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Up");
    self.prevSheet:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Down");
    self.prevSheet:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Disabled");
    self.prevSheet:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight", "ADD");
    self.prevSheet:SetScript("OnClick", function() self:Leaf(-1); self:Render(); end);
    self.prevSheet:SetPoint("LEFT", self.sheetSelector, "LEFT", 0, 0);


    self.nextSheet = CreateFrame("Button", nil, self.sheetSelector);
    self.nextSheet:SetWidth(32);
    self.nextSheet:SetHeight(32);
    self.nextSheet:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Up");
    self.nextSheet:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Down");
    self.nextSheet:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Disabled");
    self.nextSheet:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight", "ADD");
    self.nextSheet:SetScript("OnClick", function() self:Leaf(1); self:Render(); end);
    self.nextSheet:SetPoint("RIGHT", self.sheetSelector, "RIGHT", 0, 0);

    self.sheetSelector:SetPoint("BOTTOMLEFT", self.frame, "BOTTOMLEFT", 0, 0);
    self.sheetSelector:SetPoint("TOPRIGHT", self.frame, "BOTTOMRIGHT", 0, 32);

    self.sheetSelectorStatus = CreateFrame("Frame", nil, self.sheetSelector);
	self.sheetSelectorStatus:SetBackdrop({
        bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
        edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
        tile = true, tileSize = 16, edgeSize = 16,
        insets = { left = 3, right = 3, top = 5, bottom = 3 }
    });
    self.sheetSelectorStatus:SetAllPoints();
	self.sheetSelectorStatus:SetBackdropColor(0.1, 0.1, 0.1, 0.5);
	self.sheetSelectorStatus:SetBackdropBorderColor(0.4, 0.4, 0.4);

    self.sheetSelectorText = self.sheetSelectorStatus:CreateFontString(nil, "DIALOG");
    --for k, v in pairs(self.headerFont) do print(k); print(v); end
    self.sheetSelectorText:SetFontObject(self.headerFont);
    self.sheetSelectorText:SetAllPoints(self.sheetSelectorStatus);
    self.sheetSelectorText:SetText("");
    self.sheetSelectorText:SetTextColor(0.9,0.75,0.2);
    self.sheetSelectorText:Show();
	self.scrollFrame:SetScript("OnMouseWheel", self.scrollFrame.onMouseWheel);
    self.outerScrollframe:SetScript("OnShow", function() self.scrollFrame:onMouseWheel(0); end);
    self.scrollFrame.scrollFunc = function(caller, value)
		local viewheight = self.scrollFrame:GetHeight()
		local height = self.innerFrame:GetHeight()
		local offset

		if viewheight > height then
			offset = 0
            self.verticalScrollbar:Hide()
		else
            self.verticalScrollbar:SetMinMaxValues(0, height-viewheight)
            self.verticalScrollbar:Show()
			offset = value;
		end
		self.innerFrame:ClearAllPoints()
		self.innerFrame:SetPoint("TOPLEFT", 0, offset)
		self.innerFrame:SetPoint("TOPRIGHT", 0, offset)

		self.scrollFrame.scrollvalue = value
    end;

    self.editbox = CreateFrame("EditBox", nil, self.frame, "InputBoxTemplate");
	self.editbox:SetAutoFocus(true);
	self.editbox:SetFontObject(self.stringFont);
	self.editbox:SetScript("OnEscapePressed", function(me) me:Hide(); end);
    self.editbox:SetScript("OnHide", function(me) if ( me.caller ~= nil ) then me.caller:Show();end end);
	self.editbox:SetScript("OnEnterPressed",  function(me) if ( me.caller ~= nil ) then me.caller.sheet:SaveEditBox(me.caller, me:GetText()); me:Hide(); end end);
	self.editbox:SetTextInsets(0, 0, 3, 3);
	self.editbox:SetMaxLetters(256);
    self.editbox:Hide();
    self.hasBookUI = true;
end

function LibSpreadsheet:Hide()
    if ( self.frame ~= nil ) then
        self.frame:Hide();
    end
end


function LibSpreadsheet:Render()
    if ( self.hasBookUI == false ) then
        self:BuildFramework();
    end
    local sheet = self.selectedSheet;
    sheet:UpdateFrameSizes();
    for k, s in pairs(self.sheets) do
        if ( s.frame ~= nil ) then
            s.frame:Hide();
            s.header:Hide();
        end
    end
    if ( self.innerFrame ~= nil ) then
        self.innerFrame:Hide();
        self.header:Hide();
    end
    if ( sheet.frame == nil ) then
        sheet.frame = CreateFrame("Frame", nil, self.scrollFrame);
        sheet.frame:SetAllPoints(self.frame);
        sheet.frame.placeholder = CreateFrame("Frame", nil, sheet.frame);
        sheet.frame.placeholder:SetPoint("TOPLEFT", sheet.frame, "TOPLEFT", -1, 0);
        sheet.frame.placeholder:SetPoint("BOTTOMRIGHT", sheet.frame, "TOPRIGHT", -1, -2);
    end
    if ( sheet.header == nil ) then
        sheet.header = CreateFrame("Frame", nil, self.frame);
        sheet.header:SetPoint("TOPLEFT", self.frame, "TOPLEFT", 0, 0);
        sheet.header:SetPoint("BOTTOMRIGHT", self.frame, "TOPLEFT", 0, -20);
    end

    self.innerFrame = sheet.frame;
    self.header = sheet.header;
    self.scrollFrame:SetScrollChild(sheet.frame);
    self.verticalScrollbar:Hide();
    sheet.frame:Show();
    sheet.header:Show();

    local offset = 0;
    for key, cell in pairs (sheet.columns) do
        if ( cell.hidden == true ) then
            if ( cell.frame ~= nil ) then
                cell.frame:Hide();
            end
        else
            if ( cell.frame == nil ) then
                local cellFrame = CreateFrame("button", nil, self.header);
                cellFrame.label = cellFrame:CreateFontString(nil, "DIALOG");
                cellFrame.label:SetFontObject(self.headerFont);
                cellFrame.label:SetPoint("TOPLEFT", cellFrame, "TOPLEFT", (sheet.cellPadding), -(sheet.cellPadding));
                cellFrame.label:SetPoint("BOTTOMRIGHT", cellFrame, "BOTTOMRIGHT", -(sheet.cellPadding), (sheet.cellPadding));
                cell.frame = cellFrame;
            end
            local cellFrame = cell.frame;
            if ( cell.u ) then
                if ( cell.sort == true ) then
                    local k = key;
                    cellFrame:SetScript("OnClick",
                        function()
                            local l = true;
                            if ( sheet.callbacks["OnUserSortRequested"] ~= nil ) then
                                if ( type(sheet.callbacks["OnUserSortRequested"]) == "function" ) then
                                    l = (sheet.callbacks["OnUserSortRequested"])(LibSpreadsheet.cols[k]);
                                end
                            end
                            if ( l ) then
                                sheet:SortByColumn(k, sheet.sortDirection);
                                self:Render();
                                if ( sheet.callbacks["OnUserSortProcessed"] ~= nil ) then
                                    if ( type(sheet.callbacks["OnUserSortProcessed"]) == "function" ) then
                                        (sheet.callbacks["OnUserSortProcessed"])(LibSpreadsheet.cols[k]);
                                    end
                                end
                            end
                        end
                    );
                end
                cellFrame.label:SetText((cell.name or "??")..":");
                cellFrame.label:SetTextColor(0.9,0.75,0.2);
                cellFrame.label:SetJustifyH(cell.justify);
            end
            cellFrame:SetWidth(cell.colwidth);
            cellFrame:SetHeight(sheet.rowHeight);
            cellFrame:SetPoint("TOPLEFT", self.header, "TOPLEFT", offset, 0);
            offset = offset + cell.colwidth + 2;
        end
    end

    local offsetY = sheet.cellSpacing/2;
    local prevRow = sheet.frame.placeholder;
    local rowTable = sheet.rows;
    if ( sheet.hasGroups or sheet.hasFilters ) then rowTable = sheet.rowHeap; end
    for rkey, row in pairs (rowTable) do
        local offsetX = 0;
        if ( row.frame == nil ) then
            row.frame = CreateFrame("Frame", "Row" .. rkey, sheet.frame);
            row.frame:ClearAllPoints()
            row.frame:SetPoint("TOPLEFT", prevRow, "BOTTOMLEFT", 0,  -sheet.cellSpacing);
            row.frame:SetPoint("TOPRIGHT", prevRow, "BOTTOMRIGHT", 0, -sheet.cellSpacing);
            row.frame:SetFrameLevel(rkey*10);
            row.frame:SetHeight(sheet.rowHeight);
            row.frame.texture = row.frame:CreateTexture();
            if ( row.b ~= nil ) then
                row.frame.texture:SetTexture(row.b.r or row.b[1], row.b.g or row.b[2], row.b.b  or row.b[3], row.b.a or row.b[4]);
            end
            row.frame.texture:SetAllPoints(row.frame)
            local myPrevRow = prevRow;
        end
        if ( row.u ) then
            row.frame:ClearAllPoints()
            row.frame:SetPoint("TOPLEFT", prevRow, "BOTTOMLEFT", 0,   -sheet.cellSpacing);
            row.frame:SetPoint("TOPRIGHT", prevRow, "BOTTOMRIGHT", 0,  -sheet.cellSpacing);
            row.frame:SetHeight(sheet.rowHeight);
            if ( row.b ~= nil ) then
                row.frame.texture:SetTexture(row.b.r or row.b[1], row.b.g or row.b[2], row.b.b  or row.b[3], row.b.a or row.b[4]);
            end
            local myPrevRow = prevRow;
            row.u = false;
        end
        prevRow = row.frame;
        local i = 0;
        for key, cell in pairs(row.cells) do
            i = i + 1;
            if ( sheet.columns[i].hidden == true ) then
                if ( cell.frame ~= nil ) then
                    cell.frame:Hide();
                end
            else
                if ( cell.frame == nil ) then
                    local cellFrame = CreateFrame("Button", nil, row.frame);
                    cellFrame.label = cellFrame:CreateFontString(nil, "DIALOG");
                    cellFrame.label:SetPoint("TOPLEFT", cellFrame, "TOPLEFT", (sheet.cellPadding), -(sheet.cellPadding));
                    cellFrame.label:SetPoint("BOTTOMRIGHT", cellFrame, "BOTTOMRIGHT", -(sheet.cellPadding), (sheet.cellPadding));
                    cellFrame.tex = cellFrame:CreateTexture(nil, "OVERLAY");
                    cellFrame.tex:SetAllPoints();
                    cellFrame.tex:SetBlendMode("ADD");
                    cellFrame.back = cellFrame:CreateTexture(nil, "BACKGROUND");
                    cellFrame.back:SetAllPoints();
                    cell.frame = cellFrame;
                end
                local cellFrame = cell.frame;
                cellFrame:SetPoint("TOPLEFT", row.frame, "TOPLEFT", offsetX, -(sheet.cellSpacing/2));
                cellFrame:SetWidth(sheet.columns[i].colwidth);
                cellFrame:SetHeight(sheet.rowHeight);
                cellFrame.column = i;
                cellFrame.row = rkey;
                cellFrame.sheet = sheet;
                cellFrame.cell = cell;
                offsetX = offsetX + sheet.columns[i].colwidth + sheet.cellPadding;

                if ( cell.u ) then
                    cellFrame.label:SetJustifyH(sheet.columns[i].justify);
                    cellFrame.label:SetTextColor(sheet.columns[i].c[1], sheet.columns[i].c[2], sheet.columns[i].c[3], 1);
                    cellFrame.back:SetTexture(sheet.columns[i].b[1], sheet.columns[i].b[2], sheet.columns[i].b[3], sheet.columns[i].b[4] or 1);
                    if ( cell.c ~= nil ) then
                        cellFrame.label:SetTextColor(cell.c.r or cell.c[1],  cell.c.g or cell.c[2],cell.c.b or cell.c[3], cell.c[4] or cell.c.a or 1);
                    end
                    if ( cell.b ~= nil ) then
                        cellFrame.back:SetTexture(cell.b[1],cell.b[2], cell.b[3], cell.b[4] or 1);
                    end
                    if ( sheet.isInteractive or sheet.columns[i].interactive ) then
                        cellFrame:SetScript("OnMouseDown", function(caller, event) sheet:StartSelect(cellFrame, event); end);
                        cellFrame:SetScript("OnMouseUp", function(caller, event) sheet:EndSelect(cellFrame, event); end);
                        cellFrame:SetScript("OnEnter", function(caller, event) sheet:UpdateSelect(cellFrame); end);
                        cellFrame:SetScript("OnLeave", function(caller, event) sheet:UpdateSelect(cellFrame); end);
                        cellFrame:SetScript("OnDoubleclick", function(caller, event) sheet:ShowEditBox(cellFrame); end);
                    end
                    if ( sheet.columns[i].onClick ~= nil ) then
                        local func = sheet.columns[i].onClick;
                        cellFrame:SetScript("OnClick", function(caller, event) func(cell.id, cell.data, cellFrame); end);

                    end

                    if ( sheet.columns[i].type == "number" ) then
                        cellFrame.label:SetFontObject(self.numberFont);
                        cell.data = cell.data or 0;
                        local extra = "";
                        if ( sheet.columns[i].decimals or 0 > 0 ) then
                            local extra = string.sub(string.format ( ("%%.%uf"):format(sheet.columns[i].decimals or 0), cell.data - math.floor(cell.data)), 3);
                        end
                        local number = ("%.0f"):format(cell.data or 0);
                        local left,num,right = string.match(number,'^([^%d]*%d)(%d+)(.-)$');
                        local prettyNum = (left and left..(num:reverse():gsub('(%d%d%d)','%1,'):reverse()) or number);
                        cellFrame.label:SetText(string.format("%s%s%s", prettyNum, extra, (sheet.columns[i].suffix or "")));
                        if ( sheet.columns[i].colorRange ~= nil ) then
                            if ( sheet.columns[i].minval ~= nil and sheet.columns[i].maxval ~= nil ) then
                                local colors = table.maxn(sheet.columns[i].colorRange);
                                if ( sheet.columns[i].maxval ~= nil and (cell.data or 0) > sheet.columns[i].maxval ) then
                                    cell.data = sheet.columns[i].maxval;
                                end
                                local where = ( ( (cell.data - sheet.columns[i].minval)  / (sheet.columns[i].maxval - sheet.columns[i].minval)) * (colors-1) ) + 1;
                                local p, n = math.floor(where), math.ceil(where);
                                local r = (where - p) * 100;
                                if (  sheet.columns[i].colorRange[p] == nil ) then
                                    where = ((colors-1)/2) + 1
                                    p, n = math.floor(where), math.ceil(where);
                                    r = 0;
                                end
                                local color = {
                                    (  ( sheet.columns[i].colorRange[p][1] * (100-r) ) + ( sheet.columns[i].colorRange[n][1] * (r) ) ) / 100,
                                    (  ( sheet.columns[i].colorRange[p][2] * (100-r) ) + ( sheet.columns[i].colorRange[n][2] * (r) ) ) / 100,
                                    (  ( sheet.columns[i].colorRange[p][3] * (100-r) ) + ( sheet.columns[i].colorRange[n][3] * (r) ) ) / 100
                                    };
                                cellFrame.label:SetTextColor(color[1], color[2], color[3]);
                                if ( sheet.columns[i].dominant ) then
                                    for k, v in pairs(sheet.rows[rkey].cells) do
                                        v.color = {color[1], color[2], color[3]};
                                        if ( v.frame ) then
                                            v.frame.label:SetTextColor(color[1], color[2], color[3]);
                                        end
                                    end
                                end
                            end
                        end
                    elseif ( sheet.columns[i].type == "string" ) then
                        cellFrame.label:SetFontObject(self.stringFont);
                        cellFrame.label:SetText(cell.data or "");
                    elseif ( sheet.columns[i].type == "boolean" ) then
                        cellFrame.label:SetFontObject(self.stringFont);
                        if ( cell.data == true ) then cellFrame.label:SetText("true");
                        else cellFrame.label:SetText("false"); end
                    end
                end
            end
        end
        offsetY = offsetY + sheet.rowHeight + 2 + sheet.cellSpacing;
    end
    self.offsetY = offsetY;
    self.scrollFrame:SetPoint("TOPLEFT", self.header, "BOTTOMLEFT", 0, 0);
    self.scrollFrame:SetPoint("BOTTOMRIGHT", self.frame, "BOTTOMRIGHT", -16, 16);
    self.verticalScrollbar:SetMinMaxValues(0, offsetY);
    self.verticalScrollbar:SetValue(0);
    self.innerFrame:SetHeight(offsetY);
    self.scrollFrame:scrollFunc(0);
    self.scrollFrame:onMouseWheel(0);
    self.sheetSelector:Hide();
    if ( self.showSelector == true ) then
        self:Leaf(0);
        self.scrollFrame:SetPoint("BOTTOMRIGHT", self.frame, "BOTTOMRIGHT", -16, 48);
        self.sheetSelector:Show();
    end
    return self.frame;
end

function LibSpreadsheet:SetCellSpacing(size)
    if ( type(size) ~= "number" ) then error(("%s: Error in setting cell spacing; Expected nuber, got %s!"):format(SPREADSHEET, type(number))); end
    self.cellSpacing = number;
end

function LibSpreadsheet:SetCellPadding(size)
    if ( type(size) ~= "number" ) then error(("%s: Error in setting cell padding; Expected nuber, got %s!"):format(SPREADSHEET, type(number))); end
    self.cellPadding = number;
end

function LibSpreadsheet:SetInteractive(toggle)
    if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting interactive mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
    self.isInteractive = toggle;
end

function LibSpreadsheet:SetColorRange(column, ...)
    column = self:ExtractLocation(column);
    if ( self.columns[column] == nil ) then error(("%s: Error while setting color range for column %s: No such column!"):format(SPREADSHEET, LibSpreadsheet.cols[column])); end
    self.columns[column].colorRange = {};
    for i = 1, select("#", ...) do
		local color = select(i, ...);
        table.insert(self.columns[column].colorRange, color);
    end
end


function LibSpreadsheet:Leaf(direction)
    local c = 1;
    for k, v in pairs(self.sheets) do
        if ( v.title == self.selectedSheet.title ) then
            c = k;
            break;
        end
    end
    if ( direction == 1 ) then
        c = c + 1;
        if ( self.sheets[c] ~= nil ) then
            self.selectedSheet = self.sheets[c];
        end
    elseif ( direction == -1 ) then
        c = c - 1;
        if ( self.sheets[c] ~= nil ) then
            self.selectedSheet = self.sheets[c];
        end
    end
    if ( c < 2 ) then
        self.prevSheet:Disable();
    else
        self.prevSheet:Enable();
    end
    if ( c == table.maxn(self.sheets) ) then
        self.nextSheet:Disable();
    else
        self.nextSheet:Enable();
    end
    self.sheetSelectorText:SetText(self.title .. ": " .. self.selectedSheet.title);
    self.verticalScrollbar:SetValue(0);
end

function LibSpreadsheet:SetVisibleSelection(fromCol, fromRow, toCol, toRow)
    if ( type(fromCol) == "number" ) then
        self.selectStartColumn = fromCol;
        self.selectStartRow = fromRow;
    elseif ( type(fromCol) == "string" ) then
        local c, r = self:ExtractLocation(fromCol);
        if ( type(r) == "number" ) then
            self.selectStartColumn = c;
            self.selectStartRow = r;
            toRow = toCol;
            toCol = fromRow;
        else
            self.selectStartColumn = c;
            self.selectStartRow = fromRow;
        end
    end
    if ( type(toCol) == "number" ) then
        self.selectEndColumn = toCol;
        self.selectEndRow = toRow;
    elseif ( type(toCol) == "string" ) then
        local c, r = self:ExtractLocation(toCol);
        if ( type(r) == "number" ) then
            self.selectEndColumn = c;
            self.selectEndRow = r;
        else
            self.selectEndColumn = c;
            self.selectEndRow = toRow;
        end
    end
    self.selecting = true;
    self:UpdateSelect(nil);
    self.selecting = false;
end


function LibSpreadsheet:StartSelect(caller, event)
    self.selecting = true;
    self.selectStartColumn = caller.column;
    self.selectStartRow = caller.row;
    if ( event == "RightButton" ) then
        self.selectStartColumn = nil;
        self.selectStartRow = nil;
        self.selectEndColumn = nil;
        self.selectEndRow = nil;
    end
    self:UpdateSelect(caller);
end

function LibSpreadsheet:EndSelect(caller, event)
    self.selecting = false;
    if ( event == "LeftButton" ) then
        if not ( self.selectStartColumn == self.selectEndColumn and self.selectStartRow == self.selectEndRow ) then
            if ( self.callbacks['OnSelectionChange'] ~= nil ) then
                if ( type(self.callbacks['OnSelectionChange']) == "function" ) then
                    (self.callbacks['OnSelectionChange'])(math.min(self.selectStartColumn, self.selectEndColumn), math.min(self.selectStartRow, self.selectEndRow), math.max(self.selectStartColumn, self.selectEndColumn), math.max(self.selectStartRow, self.selectEndRow));
                end
            end
        end
    end
end

function LibSpreadsheet:UpdateSelect(caller)
    local rowTable = self.rows;
    if ( self.hasGroups or self.hasFilters ) then rowTable = self.rowHeap; end
    if ( self.selecting or false == true ) then
        if ( caller ~= nil ) then
            self.selectEndColumn = caller.column;
            self.selectEndRow = caller.row;
        end
        for x = 1, table.maxn(rowTable) do
            for y = 1, table.maxn(self.columns) do
                rowTable[x].cells[y].frame.tex:SetTexture(0,0,0,0);
            end
        end
        if ( self.selectEndRow == nil ) then self.selectEndRow  = self.selectStartRow ; end
        if ( self.selectStartRow ~= nil ) then
            for x = math.min(self.selectStartRow, self.selectEndRow), math.max(self.selectStartRow, self.selectEndRow) do
                for y = math.min(self.selectStartColumn, self.selectEndColumn), math.max(self.selectStartColumn, self.selectEndColumn) do
                    rowTable[x].cells[y].frame.tex:SetTexture(0.3,0.2,1,0.25);
                end
            end
        end
    end
end

function LibSpreadsheet:HideColumn(...)
	for i = 1, select("#", ...) do
		local column = self:ExtractLocation(select(i, ...));
        if ( self.columns[column] ~= nil ) then
            self.columns[column].hidden = true;
        end
    end
end

function LibSpreadsheet:ShowColumn(...)
	for i = 1, select("#", ...) do
		local column = self:ExtractLocation(select(i, ...));
        if ( self.columns[column] ~= nil ) then
            self.columns[column].hidden = false;
        end
    end
end

function LibSpreadsheet:SetSheetCallback(event, func)
    local found = false;
    for k,v in pairs(LibSpreadsheet.callbackEvents) do
        if ( v == event ) then
            found = true;
            break;
        end
    end
    if ( found == true ) then
        if ( type(func) == "function" ) then
            self.callbacks[event] = func;
        else
            error(("%s: Error in SetCallback(): expected function, got %s!"):format(SPREADSHEET, type(func)));
        end
    else
        error(("%s: Error in SetCallback(): No such event, '%s'!"):format(SPREADSHEET, event));
    end
end

function LibSpreadsheet:ShowEditBox(cellFrame)
    local ret = true;
    if ( self.parent().editbox:IsVisible() ) then
        self.parent().editbox:Hide();
    end
    if ( self.callbacks['OnUserInputRequested'] ~= nil ) then
        if ( type(self.callbacks['OnUserInputRequested']) == "function" ) then
            ret = (self.callbacks['OnUserInputRequested'])(cellFrame.col, cellFrame.row, cellFrame.cell.id, cellFrame.cell.data);
        end
    end
    if ( ret ) then
        self.parent().editbox:ClearAllPoints();
        self.parent().editbox.caller = cellFrame;
        self.parent().editbox:SetAllPoints(cellFrame);
        self.parent().editbox:SetHeight(15);
        self.parent().editbox:Show();
        cellFrame:Hide();
    end
end

function LibSpreadsheet:SaveEditBox(cellFrame, text)
    local ret = true;
    if ( self.callbacks['OnUserInputProcessed'] ~= nil ) then
        if ( type(self.callbacks['OnUserInputProcessed']) == "function" ) then
            ret = (self.callbacks['OnUserInputProcessed'])(cellFrame.col, cellFrame.row, cellFrame.cell.id, cellFrame.cell.data, text);
        end
    end
    if ( ret ) then
        cellFrame.cell.data = text;
        cellFrame.cell.dataChanged = true;
        self.parent():Render();
    end
    self.parent().editbox:ClearAllPoints();
    self.parent().editbox:SetText("");
end


--[[
LibSpreadsheet-1.0 EXAMPLES
]]--

function LibSpreadTest()
    local lss = LibStub("LibSpreadsheet-1.0", true);                    -- Load the lib
    local book = lss:Book("LibSpreadsheet examples");                   -- Make a new book
    local sheet = book:Add("My stuff");                                 -- Add a sheet to the book
    book:StrictMode(true);
    sheet.isInteractive = true;
    book:ShowSelector(true);
    sheet:AddColumn(
        {name="Name", sort=true, background={0.7,0.9,0.8,0.3}},
        "Class",
        {name="Level", type="number", minval=1, maxval=80, sort=true, background={1,0.9,0.2,0.3}},
        {name="Score", type="number", decimals=2, suffix="%", sort=true},
        {name="Gnomish", type="boolean", color={0,1,1}}
        );  -- Set some headers
    sheet:AddRow("Apricot", "Warrior", 79, 95.5, true);                     -- Add some data
    sheet:AddRow("Bumblebee", "|cffFF4499Priest|r", 27, 82.5, false);                      -- Add another row
    sheet:AddRow("Charlie", "Gnome?", 50, 23, true);                     -- Add some data
    sheet:AddRow("Doris", "Warriorwarriorwarrior", 60, 31.34, true);                     -- Add some data
    local y = sheet:AddRow("Evelyn", "|cffFF4499Priest|r", 80, 45.7, false);                      -- Add another row
    sheet:SetRow(y, {background={1,0.4,0.2,0.2}});
    sheet:AddRow("Frank", "Gnome....", 1, 67.5, true);                     -- Add some data
    sheet:AddRow("George", "Warrior", 40, 8.4, true);                     -- Add some data
    sheet:AddRow("Henry", "|cffFF4499Priest|r", 75, 2.5, false);                      -- Add another row
    sheet:AddRow("Ian", "Gnome....", 12, 11, true);                     -- Add some data
    sheet:SetColorRange("C", {1,0,0}, {1,1,0}, {0,1,0});
    sheet:GroupBy("B");
    sheet = book:Add("Other stuff");
    sheet:AddColumn("Eeny", "Meeny", "Miny", "Mo");
    sheet:AddRow("Meeny", "Miny", "Mo", "Eeny");
    sheet:AddRow("Miny", "Mo", "Eeny", "Meeny");
    sheet:AddRow("Mo", "Eeny", "Meeny", "Miny");

    sheet = book:Select("My stuff");
    sheet:SetCellValue("B", 2, "Mooo");
    sheet:SetCellValue(3, 3, 77);
	if ( UIParent ) then
        local frame = book:Render();                                        -- Create the UI
        local window = CreateFrame("Frame", nil, UIParent, "DialogBoxFrame");
        window:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
        window:SetWidth(400);
        window:SetHeight(250);
        window:RegisterForDrag("LeftButton");
        window:SetScript("OnDragStart",function()  window:StartMoving(); end);
        window:SetScript("OnDragStop", function()  window:StopMovingOrSizing(); end);
        window:SetMovable(true);
        window:EnableMouse(true);
        frame:SetParent(window);
        frame:SetPoint("TOPLEFT", window, "TOPLEFT", 20, -20);
        frame:SetPoint("BOTTOMRIGHT", window, "BOTTOMRIGHT", -20, 50);
        frame:Show();
        window:Show();
	end
end
